import importlib
from pathlib import Path

import pytest

from .common import create_fake_plugin

dummy_plugin_py = """
from kaizo.plugins import Plugin

class MyPlugin(Plugin):
    pass
"""

missing_config = """
plugins:
  nonexistent:
    source: MyPlugin
"""

invalid_plugin_config = """
plugins: not_a_dict
"""

missing_source_config = """
plugins:
  missing:
    args:
      z: 10
"""

invalid_plugin_module_config = """
plugins:
  invalid: 123
"""

bad_plugin_config = """
plugins:
  bad: MyPlugin
"""

wrong_plugin_config = """
plugins:
  wrong: MyPlugin
"""
wrong_plugin_py = """
class MyPlugin:
    pass
"""

using_invalid_plugin_config = """
invalid:
  module: plugin
  source: invalid
"""

plugin_not_found_config = """
plugins:
  dummy: MyPlugin
not_found:
  module: plugin
  source: not_found
"""


def test_plugin_missing_module(tmp_path: Path) -> None:
    kaizo = importlib.import_module("kaizo")

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(missing_config)

    with pytest.raises(ImportError):
        kaizo.ConfigParser(cfg_file)


def test_plugin_invalid_type(tmp_path: Path) -> None:
    kaizo = importlib.import_module("kaizo")

    cfg = tmp_path / "cfg.yml"
    cfg.write_text(invalid_plugin_config)

    with pytest.raises(TypeError):
        kaizo.ConfigParser(cfg)


def test_plugin_missing_source_key(tmp_path: Path) -> None:
    kaizo = importlib.import_module("kaizo")

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(missing_source_config)

    with pytest.raises(ValueError, match="source is required for missing plugin"):
        kaizo.ConfigParser(cfg_file)


def test_plugin_missing_source_attr(tmp_path: Path) -> None:
    create_fake_plugin(tmp_path, "bad", body="")
    kaizo = importlib.import_module("kaizo")

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(bad_plugin_config)

    with pytest.raises(AttributeError):
        kaizo.ConfigParser(cfg_file)


def test_plugin_invalid_module(tmp_path: Path) -> None:
    kaizo = importlib.import_module("kaizo")

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(invalid_plugin_module_config)

    with pytest.raises(TypeError):
        kaizo.ConfigParser(cfg_file)


def test_plugin_not_subclass(tmp_path: Path) -> None:
    create_fake_plugin(tmp_path, "wrong", body=wrong_plugin_py)
    kaizo = importlib.import_module("kaizo")

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(wrong_plugin_config)

    with pytest.raises(TypeError):
        kaizo.ConfigParser(cfg_file)


def test_using_invalid_plugin(tmp_path: Path) -> None:
    kaizo = importlib.import_module("kaizo")

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(using_invalid_plugin_config)

    parser = kaizo.ConfigParser(cfg_file)

    with pytest.raises(ValueError, match="plugins are not given"):
        parser.parse()


def test_plugin_not_found(tmp_path: Path) -> None:
    create_fake_plugin(tmp_path, "dummy", body=dummy_plugin_py)
    kaizo = importlib.import_module("kaizo")

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(plugin_not_found_config)

    parser = kaizo.ConfigParser(cfg_file)

    with pytest.raises(ValueError, match="plugin not_found not found"):
        parser.parse()



from pathlib import Path

from kaizo import ConfigParser

X1 = 5
X2 = 6
X3 = 7

main_py = """
def calc(x):
    return x
"""

main_config = f"""
local: main.py
x: {X1}
y:
  module: local
  source: calc
  args:
    x: {X2}
z: {X3}
x-ref: .{{x}}
y-ref: .y.{{}}
x-y-ref: .y.{{x}}
"""

implicit_config = """
local: main.py
import:
 m: main.yml
x: m.{}
y: m.y.{}
z:
  module: local
  source: calc
  args:
    - m.{}
"""

explicit_config = """
local: main.py
import:
 m: main.yml
x: m.{x}
y: m.{y}
y-x: m.y.{x}
z:
  module: local
  source: calc
  args:
    - m.{z}
"""


def test_local_ref(tmp_path: Path) -> None:
    module = tmp_path / "main.py"
    module.write_text(main_py)

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(main_config)

    parser = ConfigParser(cfg_file)
    out = parser.parse()

    assert out["x-ref"] == X1
    assert out["y-ref"] == X2
    assert out["x-y-ref"] == X2


def test_module_implicit_ref(tmp_path: Path) -> None:
    module = tmp_path / "main.py"
    module.write_text(main_py)

    main_file = tmp_path / "main.yml"
    main_file.write_text(main_config)

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(implicit_config)

    parser = ConfigParser(cfg_file)
    out = parser.parse()

    assert out["x"] == X1
    assert out["y"] == X2
    assert out["z"] == X3


def test_module_explicit_ref(tmp_path: Path) -> None:
    module = tmp_path / "main.py"
    module.write_text(main_py)

    main_file = tmp_path / "main.yml"
    main_file.write_text(main_config)

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(explicit_config)

    parser = ConfigParser(cfg_file)
    out = parser.parse()

    assert out["x"] == X1
    assert out["y"] == X2
    assert out["y-x"] == X2
    assert out["z"] == X3



from pathlib import Path

from kaizo import ConfigParser

X = 5

config = """
call_me:
  module: math
  source: pow
  call: true
  args:
    - .{injected}
    - 2
"""


def test_kwargs_injection(tmp_path: Path) -> None:
    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(config)

    parser = ConfigParser(cfg_file, kwargs={"injected": X})
    out = parser.parse()
    assert out["call_me"] == X**2



from pathlib import Path

from kaizo import ConfigParser
from kaizo.utils import DictEntry

A = 10
B = "hello"
C = 3.5
PATH = "./local/path"

dict_config = f"""
a: {A}
b: {B}
c: {C}
"""

lst_config = """
lst:
  - 1
  - 2
  - 3
"""

path_config = f"""
path: {PATH}
"""


def test_basic_scalars(tmp_path: Path) -> None:
    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(dict_config)

    parser = ConfigParser(cfg_file)
    res = parser.parse()

    assert isinstance(res, DictEntry)

    assert res["a"] == A
    assert res["b"] == B
    assert res["c"] == C


def test_list_parsing(tmp_path: Path) -> None:
    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(lst_config)

    parser = ConfigParser(cfg_file)
    res = parser.parse()

    assert isinstance(res, DictEntry)

    lst_entry = res["lst"]

    assert list(lst_entry) == [1, 2, 3]


def test_basic_path(tmp_path: Path) -> None:
    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(path_config)

    parser = ConfigParser(cfg_file)
    res = parser.parse()

    assert isinstance(res, DictEntry)

    assert res["path"] == PATH



import importlib
import shutil
import sys
from pathlib import Path


def create_fake_plugin(tmp_path: Path, name: str, body: str) -> Path:
    real_kaizo = importlib.import_module("kaizo")

    real_kaizo_path = Path(real_kaizo.__file__).parent

    dest_path = tmp_path / "kaizo"

    shutil.copytree(real_kaizo_path, dest_path)

    plugins_dir = dest_path / "plugins"
    plugins_dir.mkdir(parents=True, exist_ok=True)

    plugin_file = plugins_dir / f"{name}.py"
    plugin_file.write_text(body)

    if str(tmp_path) in sys.path:
        sys.path.remove(str(tmp_path))
    sys.path.insert(0, str(tmp_path))

    for mod in list(sys.modules):
        if mod == "kaizo" or mod.startswith("kaizo."):
            del sys.modules[mod]

    importlib.invalidate_caches()

    return plugin_file



from collections.abc import Callable
from copy import deepcopy
from functools import lru_cache
from typing import Generic, TypeVar

R = TypeVar("R")


class FnWithKwargs(Generic[R]):
    fn: Callable[..., R]
    args: tuple | None
    kwargs: dict[str] | None

    def __init__(
        self,
        fn: Callable[..., R],
        args: tuple | None = None,
        kwargs: dict[str] | None = None,
        *,
        cache: bool = False,
    ) -> None:
        if cache:
            self.fn = lru_cache()(fn)
        else:
            self.fn = fn

        if kwargs is None:
            kwargs = {}

        self.args = args
        self.kwargs = kwargs

    def __call__(self, *args, **kwargs) -> R:
        call_kwargs = deepcopy(self.kwargs)
        call_kwargs.update(kwargs)

        if self.args is not None:
            args = self.args

        return self.fn(*args, **call_kwargs)

    def update(self, **kwargs) -> None:
        self.kwargs.update(kwargs)
